class MissingResourcesError(Exception):
    pass


class MissingRequirementError(Exception):
    pass


class RequirementVersionError(Exception):
    pass


class NoContractionException(Exception):
    pass


class InvalidRangeException(Exception):
    pass


class BasisNotFoundError(Exception):
    pass
